import subprocess
import click

import requests
from .ChatGPT import Chatbot
import os
import sys
from rich.live import Live
from rich.markdown import Markdown
from .code_execution import *
import yaml
from pathlib import Path

sys.stdin.reconfigure(encoding='utf-8')

@click.command()
def hello():
    """Say hello"""
    click.echo('Hello, world!')

@click.group()
def gpterminal():
    """Figlu command line tool"""
    # 默认调用bash函数
    pass


def chatgpt_explanation(api_key, base_url, language, error_message, model, maxtoken):
    query = construct_query(language, error_message)
    chatbot = Chatbot(api_key=api_key, base_url=base_url, engine=model, max_tokens=maxtoken)
    res = chatbot.ask_stream(query)
    return res


def get_api_key():
    # 确定配置文件的存储位置
    config_path = Path.home() / ".config" / "GPTerminal" / "config.yaml"

    # 检查配置文件是否存在，如果不存在则提示用户创建
    if not config_path.is_file():
        print(f"Config file not found at {config_path}.")
        # 创建配置文件所在的目录
        config_path.parent.mkdir(parents=True, exist_ok=True)
        # 创建配置文件
        config_path.touch()
        while True:
            api_key = input("Please enter your OpenAI API key: ").strip()
            if api_key:
                break
            else:
                print("API key cannot be empty. Please try again.")
        while True:
            base_url = input("Please enter the OpenAI API base URL (or leave blank): ").strip()
            if base_url or base_url == "":
                break
            else:
                print("Base URL cannot be empty. Please try again.")

        # 将 API key 和 base URL 写入配置文件
        with open(config_path, "w") as f:
            yaml.safe_dump({"api_key": api_key, "base_url": base_url or None}, f)
    else:
        #读配置文件中的 API key 和 base URL
        with open(config_path) as f:
            config = yaml.safe_load(f)
            api_key = config.get("api_key", None)
            base_url = config.get("base_url", None)

    # 如果没有读取到 API key，则引发异常
    if not api_key:
        raise Exception("Please enter OpenAI API key in the config file.")

    return base_url, api_key




@click.command()
@click.option("--model", default="gpt-3.5-turbo", help="Specify which GPT model to use")
@click.option("--maxtoken", default=4096, help="Maximum number of tokens in a single prompt")
def chat(model, maxtoken):
    """Chat with GPT-3 or GPT-4"""
    base_url, api_key = get_api_key()
    bot = Chatbot(api_key=api_key, base_url=base_url, engine=model, max_tokens=maxtoken)
    while True:
        text = input("You: ")
        if text.strip() == "exit":
            break
        response = bot.ask_stream(text)
        md = Markdown("")
        with Live(md, auto_refresh=False) as live:
            tmp = ""
            for r in response:
                tmp += r
                md = Markdown(tmp)
                live.update(md, refresh=True)
                

@click.command()
@click.option("--model", default="gpt-3.5-turbo", help="Specify which GPT model to use")
@click.option("--maxtoken", default=4096, help="Maximum number of tokens in a single prompt")
def git(model, maxtoken):
    """问问关于git的问题"""
    print("这是一个git命令行助手,你可以通过这个助手来学习git命令")
    # use openai to generate git scripts
    base_url, api_key = get_api_key()
    bot = Chatbot(api_key=api_key, base_url=base_url, system_prompt="""
    你是一个很好的git命令行助手，你可以帮助我更好的使用git, 
    根据我的问题,你可以给我提供一些git命令
    """, engine=model, max_tokens=maxtoken)
    while True:
        text = input("请输入你的问题: ")
        if text.strip() == "exit":
            break
        response = bot.ask_stream(text)
        print("=" * 80)
        tmp = ""
        for r in response:
            tmp += r
            print(r, end='', flush=True)
        print()
        print("=" * 80)
            
            
            
@click.command()
@click.argument('args', nargs=-1)
@click.option("--model", default="gpt-3.5-turbo", help="Specify which GPT model to use")
@click.option("--maxtoken", default=4096, help="Maximum number of tokens in a single prompt")
@click.option("--lines", default=100, help="Number of lines before and after the error to capture as context")
def run(args, model, maxtoken, lines):
    """
    接收不固定参数的命令，并执行它们。如果发生错误，将捕获并显示错误输出及其上下文。
    """
    try:
        base_url, api_key = get_api_key()
        language = get_language(args)
        # 将参数元组转换为列表，并将其传递给subprocess.run函数
        result = subprocess.run(
            list(args), stderr=subprocess.PIPE, stdout=subprocess.PIPE, text=True, check=True
        )
    except subprocess.CalledProcessError as e:
        # 找出可能涉及的文件
        source_files = [arg for arg in args if os.path.isfile(arg)]
        file_contents = []

        for file in source_files:
            # 检查文件字数是否超过2000
            num_chars = os.stat(file).st_size
            if num_chars <= 2000:
                with open(file, 'r', encoding='utf-8') as f:
                    content = f.read()
            else:
                with open(file, 'r', encoding='utf-8') as f:
                    head_lines = [next(f, '') for _ in range(lines)]
                with open(file, 'r', encoding='utf-8') as f:
                    tail_lines = [line for _, line in zip(range(lines), f.readlines())]
                    

                content = "".join(head_lines) + "\n...\n" + "".join(tail_lines)

            file_contents.append(content)

        # 获取错误输出
        error_output = e.stderr
        
        # 打印错误输出
        click.echo(error_output)

        # 将文件内容与错误输出合并
        total_input = f"Error Output:\n{error_output}\n\n"
        for i, content in enumerate(file_contents):
            total_input += f"File Content ({source_files[i]}):\n{content}\n\n"

        # 使用GPT模型处理整个输入
        with LoadingMessage():
            response = chatgpt_explanation(api_key, base_url, language, total_input, model, maxtoken)

        # 使用Rich库将结果显示为Markdown
        md = Markdown("")
        with Live(md, auto_refresh=True) as live:
            tmp = ""
            for r in response:
                tmp += r
                md = Markdown(tmp)
                live.update(md, refresh=True)
    else:
        click.echo(result.stdout)
    

gpterminal.add_command(hello)
gpterminal.add_command(chat)
gpterminal.add_command(git)
gpterminal.add_command(run)

if __name__ == '__main__':
    gpterminal()



# Standard library
import os
from queue import Queue
from subprocess import PIPE, Popen
from threading import Thread

# Local
# from DXRExplain.utilities.parsers import get_code_exec_command, get_error_message

GRAY = "\033[37m"
END = "\033[0m"

# Standard library
import re


######
# MAIN
######


def get_language(args):
    """
    Returns the language a file is written in.
    """

    file_path = args[1].lower()
    if file_path.endswith(".py"):
        return "python"
    elif file_path.endswith(".js"):
        return "node"
    elif file_path.endswith(".go"):
        return "go run"
    elif file_path.endswith(".rb"):
        return "ruby"
    elif file_path.endswith(".java"):
        return "javac" # Compile Java Source File
    elif file_path.endswith(".class"):
        return "java" # Run Java Class File
    else:
        return "" # Unknown language


def get_code_exec_command(args, language):
    fp_and_args = args[1:]
    if language == "java":
        fp_and_args = [arg.replace(".class", "") for arg in fp_and_args]

    return [language] + fp_and_args


def get_error_message(error, language):
    """
    Filters the stack trace from stderr and returns only the error message.
    """

    if error == '' or error is None:
        return None
    elif language == "python":
        if any(e in error for e in ["KeyboardInterrupt", "SystemExit", "GeneratorExit"]): # Non-compiler errors
            return None
        else:
            return error
    elif language == "node":
        return error.split('\n')[4][1:]
    elif language == "go run":
        return error.split('\n')[1].split(": ", 1)[1][1:]
    elif language == "ruby":
        error_message = error.split('\n')[0]
        return error_message[error_message.rfind(": ") + 2:]
    elif language == "javac":
        m = re.search(r'.*error:(.*)', error.split('\n')[0])
        return m.group(1) if m else None
    elif language == "java":
        for line in error.split('\n'):
            # Multiple error formats
            m = re.search(r'.*(Exception|Error):(.*)', line)
            if m and m.group(2):
                return m.group(2)

            m = re.search(r'Exception in thread ".*" (.*)', line)
            if m and m.group(1):
                return m.group(1)

        return None


#########
# HELPERS
#########


def read(pipe, funcs):
    """
    Reads and pushes piped output to a shared queue and appropriate lists.
    """

    for line in iter(pipe.readline, b''):
        for func in funcs:
            func(line.decode("utf-8"))

    pipe.close()


def write(get):
    """
    Pulls output from shared queue and prints to terminal.
    """

    print()
    for line in iter(get, None):
        line = line.replace("\n", "")
        print(f"{GRAY}{line}{END}")


######
# MAIN
######


def execute_code(args, language):
    """
    Executes a given command in a subshell, pipes stdout/err to the current
    shell, and returns the stderr.
    """
    command = get_code_exec_command(args, language)
    print(" ".join(command))
    process = Popen(
        command,
        cwd=None,
        shell=False,
        close_fds=True,
        stdout=PIPE,
        stderr=PIPE,
        bufsize=-1
    )

    output, errors = [], []
    pipe_queue = Queue()

    # Threads for reading stdout and stderr pipes and pushing to a shared queue
    stdout_thread = Thread(target=read, args=(process.stdout, [pipe_queue.put, output.append]))
    stderr_thread = Thread(target=read, args=(process.stderr, [pipe_queue.put, errors.append]))

    # Thread for printing items in the queue
    writer_thread = Thread(target=write, args=(pipe_queue.get,))

    # Spawns each thread
    for thread in (stdout_thread, stderr_thread, writer_thread):
        thread.daemon = True
        thread.start()

    process.wait()

    for thread in (stdout_thread, stderr_thread):
        thread.join()

    pipe_queue.put(None)

    # output = ''.join(output)
    errors = '\n'.join(errors)
    
    # 合并错误信息
    # if len(errors) > 0:
    #     errors = output + '\n' + errors

    # File doesn't exist, for java, command[1] is a class name instead of a file
    if "java" != command[0] and not os.path.isfile(command[1]):
        return None

    return get_error_message(errors, language)


def construct_query(language, error_message):
    # TODO: Create an class for mapping languages to exec commands
    language = "java" if language == "javac" else language
    language = "python" if language == "python3" else language
    language = "go" if language == "go run" else language
    print()
    query = f"Explain this {language} error message in brief and simple terms:"
    query += "\n```"
    query += f"\n{error_message}"
    query += "\n``` 给出详细的中文解决方案，以及修改建议，如果涉及到环境配置，请给出详细的环境配置步骤。"
    print()
    return query

# Standard library
import sys
import re
from itertools import cycle
from shutil import get_terminal_size
from threading import Thread
from time import sleep

# Third party
from pygments import highlight
from pygments.lexers import get_lexer_by_name
from pygments.formatters import Terminal256Formatter

CODE_IDENTIFIER = "```"
CODE_INDENT = "    "

# ASCII color codes
CYAN = "\033[36m"
RED = "\033[31m"
END = "\033[0m"
BOLD = "\033[1m"
UNDERLINE = "\033[4m"

INLINE_BY_STAR_IDENTIFIER = "\*(.*?)\*"
INLINE_BY_DASH_IDENTIFIER = "`(.*?)`"


#########
# HELPERS
#########


def handle_inline_code(text):
    replacer = lambda s: f"{BOLD}{s.group()}{END}"
    text = re.sub(INLINE_BY_STAR_IDENTIFIER, replacer, text)
    text = re.sub(INLINE_BY_DASH_IDENTIFIER, replacer, text)

    return text


def slow_print(text, delay=0.01):
    for word in text:
        sys.stdout.write(word)
        sys.stdout.flush() # Defeat buffering

        sleep(delay)


def slow_print_code(text, delay=0.0025):
    code = highlight(
        text,
        lexer=get_lexer_by_name("python"),
        formatter=Terminal256Formatter(style="gruvbox-dark")
    )
    for line in code.strip().split("\n"):
        slow_print(f"{CODE_INDENT}{line}", delay)
        print()


######
# MAIN
######


def print_help_message():
    """
    Prints usage instructions.
    """

    print(f"{BOLD}DXRExplain – Made by @shobrook{END}")
    print("Command-line tool that automatically explains your error message using ChatGPT.")
    print(f"\n\n{UNDERLINE}Usage:{END} $ DXRExplain {CYAN}[file_name]{END}")
    print(f"\n$ python3 {CYAN}test.py{END}   =>   $ DXRExplain {CYAN}test.py{END}")


def print_invalid_language_message():
    print(f"\n{RED}Sorry, DXRExplain doesn't support this file type.\n{END}")


class LoadingMessage:
    def __init__(self, timeout=0.1):
        """
        A loader-like context manager

        Args:
            desc (str, optional): The loader's description. Defaults to "Loading...".
            end (str, optional): Final print. Defaults to "Done!".
            timeout (float, optional): Sleep time between prints. Defaults to 0.1.
        """

        # self.steps = ["⢿", "⣻", "⣽", "⣾", "⣷", "⣯", "⣟", "⡿"]
        self.steps = ['-', '/', '|', '\\']
        self.message = f"{BOLD}{CYAN}正在让ChatGPT帮你查看错误{END}"
        self.end = f"{BOLD}{CYAN}🤖 ChatGPT的解释是:{END}"
        self.timeout = timeout

        self._thread = Thread(target=self._animate, daemon=True)
        self.done = False

    def start(self):
        self._thread.start()
        return self

    def stop(self):
        self.done = True
        cols = get_terminal_size((80, 20)).columns

        print(f"\r{' ' * cols}", end="", flush=True)
        print(f"\r{self.end}\n", flush=True)

    def _animate(self):
        for step in cycle(self.steps):
            if self.done:
                break

            print(f"\r{CYAN}{step}{END} {self.message}", flush=True, end="")

            sleep(self.timeout)

    def __enter__(self):
        self.start()

    def __exit__(self, exc_type, exc_value, tb):
        self.stop()


def print_error_explanation(explanation):
    for i in explanation:
        print(i, end="", flush=True)
    print()




from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="gpterminal",
    version="0.1.4",
    author="Your Name",
    author_email="your.email@example.com",
    description="A CLI tool to interact with GPT-3/GPT-4, answer Git questions, and execute commands",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/yourusername/gpterminal",
    packages=find_packages(),
    install_requires=[
        'click',
        'requests',
        'rich',
        'pathlib',
        'openai',
        'tiktoken',
        'prompt_toolkit'
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    python_requires=">=3.6",
    entry_points={
        "console_scripts": [
            "gpterminal=cli.cli:gpterminal",
        ],
    },
)
